import { Transform } from 'stream'

/**
 * Create a Transform stream which will maintain a buffer with data received from a Readable stream and write data when the buffer can be matched against the regex. It will push the whole match object (or objects when the g flag is used) returned by `/regex/.exec(buffer)`.
 * @param {!RegExp} regex The regular expression to execute.
 */
export default function createRegexTransformStream(regex) {
  let buffer = ''

  /** @suppress {checkTypes} */
  const ts = new Transform({
    transform(chunk, encoding, next) {
      let lastMatch
      let match
      buffer += chunk.toString()

      // thx stream-snitch
      // https://github.com/dmotz/stream-snitch/blob/master/index.js#L52
      // eslint-disable-next-line no-cond-assign
      while (match = regex.exec(buffer)) {
        ts.push(match)
        lastMatch = match
        if (!regex.global) break
      }
      if (lastMatch) {
        buffer = buffer.slice(lastMatch.index + lastMatch[0].length)
      }
      next()
    },
    objectMode: true,
  })

  return ts
}

export {
  makeMarkers,
  makeCutRule,
  makePasteRule,
} from './lib/markers'

export { default as Replaceable, SerialAsyncReplaceable } from './Replaceable'
export { default as SyncReplaceable } from './SyncReplaceable'

/* documentary types/rules.xml */
/**
 * @suppress {nonStandardJsDocs}
 * @typedef {_restream.Replacer} Replacer
 */
/**
 * @suppress {nonStandardJsDocs}
 * @typedef {function(this:*, ...string): string} _restream.Replacer
 */
/**
 * @suppress {nonStandardJsDocs}
 * @typedef {_restream.AsyncReplacer} AsyncReplacer
 */
/**
 * @suppress {nonStandardJsDocs}
 * @typedef {function(this:*, ...string): !Promise<string>} _restream.AsyncReplacer
 */
/**
 * @suppress {nonStandardJsDocs}
 * @typedef {_restream.Rule} Rule A replacement rule.
 */
/**
 * @suppress {nonStandardJsDocs}
 * @typedef {Object} _restream.Rule A replacement rule.
 * @prop {!RegExp} re The regular expression to match input against.
 * @prop {!(string|_restream.Replacer|_restream.AsyncReplacer)} replacement The replacement string, or the replacer function.
 */
